import asyncio
import inspect
import logging
import threading

import httpx

from . import exceptions


logger = logging.getLogger(__name__)


class Auth(httpx.Auth):
    """
    Azimuth API authenticator.
    """
    def __init__(
        self,
        base_url,
        *,
        auth_data,
        authenticator = None,
        authenticator_type = None
    ):
        assert \
            authenticator or authenticator_type, \
            "one of authenticator or authenticator_type is required"
        self.base_url = base_url.rstrip("/")
        self.authenticator = authenticator
        self.authenticator_type = authenticator_type
        self.auth_data = auth_data
        self._token = None
        # We use locks to make sure only one request refreshes the token at once
        self._sync_lock = threading.RLock()
        self._async_lock = asyncio.Lock()

    def _raise_for_status(self, response):
        try:
            response.raise_for_status()
        except httpx.HTTPStatusError as source:
            raise exceptions.APIError(source)

    def _build_authenticators_request(self):
        logger.debug("building authenticators request")
        return httpx.Request("GET", f"{self.base_url}/auth/authenticators/")

    def _handle_authenticators_response(self, response):
        self._raise_for_status(response)
        try:
            authenticator = next(
                k
                for k, v in response.json().items()
                if v["type"] == self.authenticator_type
            )
        except StopIteration:
            raise exceptions.SDKError(f"no authenticators with type '{self.authenticator_type}'")
        else:
            logger.debug(
                f"using authenticator '{authenticator}' of type '{self.authenticator_type}'"
            )
            self.authenticator = authenticator

    def _build_token_request(self):
        logger.debug("building token refresh request")
        return httpx.Request(
            "POST",
            f"{self.base_url}/auth/{self.authenticator}/token/",
            json = self.auth_data
        )
    
    def _handle_token_response(self, response):
        self._raise_for_status(response)
        logger.debug("extracting token")
        self._token = response.json()["token"]

    def _authenticate_request(self, request):
        logger.debug("applying request authentication")
        request.headers["Authorization"] = f"Bearer {self._token}"
        return request
    
    def _refresh_token(self, lock):
        token = self._token
        yield lock.acquire()
        try:
            # If the token changed in the time it took to acquire the lock,
            # there is nothing to do
            if token != self._token:
                return
            if not self.authenticator:
                response = yield self._build_authenticators_request()
                self._handle_authenticators_response(response)
            response = yield self._build_token_request()
            self._handle_token_response(response)
        finally:
            lock.release()
    
    def auth_flow(self, request, lock):
        if not self._token:
            yield from self._refresh_token(lock)
        response = yield self._authenticate_request(request)
        if response.status_code == 401:
            # Try refreshing the token
            yield from self._refresh_token(lock)
            yield self._authenticate_request(request)

    def sync_auth_flow(self, request):
        flow = self.auth_flow(request, self._sync_lock)
        try:
            action = flow.send
            to_send = None
            while True:
                try:
                    yielded_obj = action(to_send)
                except StopIteration as exc:
                    break
                try:
                    if isinstance(yielded_obj, httpx.Request):
                        to_send = yield yielded_obj
                        to_send.read()
                    else:
                        to_send = yielded_obj
                except BaseException as exc:
                    action = flow.throw
                    to_send = exc
                else:
                    action = flow.send
        finally:
            flow.close()

    async def async_auth_flow(self, request):
        flow = self.auth_flow(request, self._async_lock)
        try:
            action = flow.send
            to_send = None
            while True:
                try:
                    yielded_obj = action(to_send)
                except StopIteration as exc:
                    break
                try:
                    if inspect.isawaitable(yielded_obj):
                        to_send = await yielded_obj
                    elif isinstance(yielded_obj, httpx.Request):
                        to_send = yield yielded_obj
                        await to_send.aread()
                    else:
                        to_send = yielded_obj
                except BaseException as exc:
                    action = flow.throw
                    to_send = exc
                else:
                    action = flow.send
        finally:
            flow.close()



import re

from httpx import codes


class SDKError(Exception):
    """
    Base class for errors raised by the SDK.
    """


class APIErrorMeta(type):
    """
    Metaclass for the API error type.
    """
    __exceptions__ = {}

    def __new__(cls, name, bases, attrs, status_code = None):
        klass = super().__new__(cls, name, bases, attrs)
        klass.__status_code__ = status_code
        return klass

    def __getitem__(cls, code):
        if code not in cls.__exceptions__:
            name = re.sub(r"[^a-zA-Z0-9]", "", codes.get_reason_phrase(code))
            klass = type(name, (APIError, ), {}, status_code = code)
            cls.__exceptions__[code] = klass
        return cls.__exceptions__[code]

    def __call__(cls, source):
        status_code = source.response.status_code
        # If the status code already matches the source, we are done
        if getattr(cls, "__status_code__", None) == status_code:
            return super().__call__(source)
        else:
            return cls[status_code](source)


class APIError(SDKError, metaclass = APIErrorMeta):
    """
    Base class for Azimuth API errors.
    """
    def __init__(self, source):
        super().__init__(source.response.text)
        self.request = source.request
        self.response = source.response

    @property
    def status_code(self):
        """
        The status code of the error.
        """
        return self.response.status_code

    @property
    def reason_phrase(self):
        """
        The reason phrase for the error.
        """
        return self.response.reason_phrase



import logging
import os

import yaml

from . import auth, client, exceptions


logger = logging.getLogger(__name__)


class Configuration:
    """
    Class for an Azimuth client configuration.
    """
    def __init__(self, **kwargs):
        self._kwargs = kwargs

    def async_client(self, **kwargs):
        """
        Returns a new asynchronous client using the configuration.
        """
        logger.debug("creating async client")
        merged = self._kwargs.copy()
        merged.update(kwargs)
        return client.AsyncClient(**merged)

    def sync_client(self, **kwargs):
        """
        Returns a new synchronous client using the configuration.
        """
        logger.debug("creating sync client")
        merged = self._kwargs.copy()
        merged.update(kwargs)
        return client.SyncClient(**merged)
    
    @classmethod
    def create(
        cls,
        base_url,
        *,
        auth_data,
        authenticator = None,
        authenticator_type = None,
        **kwargs
    ):
        """
        Returns a new Azimuth client configuration with the specified auth.
        """
        return cls(
            base_url = base_url,
            auth = auth.Auth(
                base_url,
                authenticator = authenticator,
                authenticator_type = authenticator_type,
                auth_data = auth_data
            ),
            **kwargs
        )
    
    @classmethod
    def from_openstack_clouds_file(
        cls,
        base_url,
        path,
        cloud = "openstack",
        *,
        default_tenancy_id = None,
        **kwargs
    ):
        """
        Returns a new Azimuth client configuration populated from an openstack cloud-config.
        """
        logger.debug(f"loading openstack config for cloud '{cloud}' from {path}")
        with open(path) as fh:
            cloud_config = yaml.safe_load(fh)["clouds"][cloud]
        auth_type = cloud_config.get("auth_type", "v3password")
        auth = cloud_config["auth"]
        if auth_type == "v3password":
            authenticator_type = "openstack_password"
            auth_data = {
                "username": auth["username"],
                "password": auth["password"],
            }
        elif auth_type == "v3applicationcredential":
            authenticator_type = "openstack_application_credential"
            auth_data = {
                "application_credential_id": auth["application_credential_id"],
                "application_credential_secret": auth["application_credential_secret"],
            }
        else:
            raise exceptions.SDKError(f"unsupported openstack auth type '{auth_type}'")
        logger.debug(f"using authenticator type '{authenticator_type}'")
        return cls.create(
            base_url,
            authenticator_type = authenticator_type,
            auth_data = auth_data,
            default_tenancy_id = default_tenancy_id or auth.get("project_id"),
            **kwargs
        )

    @classmethod
    def from_environment(cls, base_url, **kwargs):
        """
        Returns a new Azimuth client configuration populated from the environment.
        """
        # We support the same environment variables as the openstack CLI
        os_cloud = os.environ.get("OS_CLOUD")
        if os_cloud:
            logger.debug("detected openstack environment variables")
            # Use the first file from the standard locations that exists
            for path in [
                os.environ.get("OS_CLIENT_CONFIG_FILE"),
                os.path.join(os.getcwd(), "clouds.yaml"),
                os.path.expanduser("~/.config/openstack/clouds.yaml"),
                "/etc/openstack/clouds.yaml",
            ]:
                if path:
                    if os.path.exists(path):
                        return cls.from_openstack_clouds_file(
                            base_url,
                            path,
                            os_cloud,
                            **kwargs
                        )
                    else:
                        logger.warn(f"file {path} does not exist")
            else:
                raise exceptions.SDKError("no openstack config file in standard locations")
        else:
            raise exceptions.SDKError("no suitable authentication found in environment")



import logging

import httpx

from easykube import flow, rest

from . import exceptions


logger = logging.getLogger(__name__)


class Resource(rest.Resource):
    """
    Resource class for the Azimuth API.
    """
    def _prepare_path(self, id = None, params = None):
        # Make sure that paths have a trailing slash to avoid having to handle 301s
        path, params = super()._prepare_path(id, params)
        if not path.endswith("/"):
            path = path + "/"
        return path, params


class BaseClient:
    """
    Base class for sync and async clients.
    """
    def __init__(self, *, default_tenancy_id = None, **kwargs):
        # Add the JSON header to all requests by default
        kwargs.setdefault("headers", {}).setdefault("Content-Type", "application/json")
        # Set longer default timeouts as some responses take a little while
        kwargs.setdefault("timeout", httpx.Timeout(5.0, read = 30.0))
        # Follow redirects by default
        kwargs.setdefault("follow_redirects", True)
        super().__init__(**kwargs)
        self._default_tenancy_id = default_tenancy_id

    @flow.flow
    def raise_for_status(self, response):
        # Convert response errors into ApiErrors for better messages
        try:
            yield super().raise_for_status(response)
        except httpx.HTTPStatusError as source:
            raise exceptions.APIError(source)

    @flow.flow
    def _init(self):
        """
        Perform any asynchronous initialisation that is required.
        """
        if not self._default_tenancy_id:
            logger.debug("discovering default tenancy")
            # Use the first tenancy as the default tenancy
            default_tenancy = yield self.tenancies().first()
            if default_tenancy:
                self._default_tenancy_id = default_tenancy.id
        if self._default_tenancy_id:
            logger.debug(f"using {self._default_tenancy_id} as default tenancy")

    def switch_tenancy(self, tenancy_id):
        """
        Switch to the specified tenancy.
        """
        self._default_tenancy_id = tenancy_id

    def tenancies(self):
        """
        Returns a REST resource for interacting with tenancies.
        """
        logger.debug("creating resource for tenancies")
        return Resource(self, "tenancies", prefix = "/api")
    
    def _tenancy_resource(self, resource, tenancy_id = None):
        tenancy_id = tenancy_id or self._default_tenancy_id
        if not tenancy_id:
            raise exceptions.SDKError("unable to detect default tenancy")
        logger.debug(f"creating resource for {resource} in tenancy {tenancy_id}")
        return Resource(self, resource, prefix = f"/api/tenancies/{tenancy_id}")

    def images(self, tenancy_id = None):
        """
        Returns a REST resource for interacting with the images for a tenancy.
        """
        return self._tenancy_resource("images", tenancy_id)

    def sizes(self, tenancy_id = None):
        """
        Returns a REST resource for interacting with the sizes for a tenancy.
        """
        return self._tenancy_resource("sizes", tenancy_id)

    def volumes(self, tenancy_id = None):
        """
        Returns a REST resource for interacting with the volumes for a tenancy.
        """
        return self._tenancy_resource("volumes", tenancy_id)

    def external_ips(self, tenancy_id = None):
        """
        Returns a REST resource for interacting with the external IPs for a tenancy.
        """
        return self._tenancy_resource("external_ips", tenancy_id)

    def machines(self, tenancy_id = None):
        """
        Returns a REST resource for interacting with the machines for a tenancy.
        """
        return self._tenancy_resource("machines", tenancy_id)

    def cluster_types(self, tenancy_id = None):
        """
        Returns a REST resource for interacting with the CaaS cluster types for a tenancy.
        """
        return self._tenancy_resource("cluster_types", tenancy_id)
    
    def clusters(self, tenancy_id = None):
        """
        Returns a REST resource for interacting with the CaaS clusters for a tenancy.
        """
        return self._tenancy_resource("clusters", tenancy_id)

    def kubernetes_cluster_templates(self, tenancy_id = None):
        """
        Returns a REST resource for interacting with the Kubernetes cluster templates for a tenancy.
        """
        return self._tenancy_resource("kubernetes_cluster_templates", tenancy_id)

    def kubernetes_clusters(self, tenancy_id = None):
        """
        Returns a REST resource for interacting with the Kubernetes clusters for a tenancy.
        """
        return self._tenancy_resource("kubernetes_clusters", tenancy_id)

    def kubernetes_app_templates(self, tenancy_id = None):
        """
        Returns a REST resource for interacting with the Kubernetes app templates for a tenancy.
        """
        return self._tenancy_resource("kubernetes_app_templates", tenancy_id)

    def kubernetes_apps(self, tenancy_id = None):
        """
        Returns a REST resource for interacting with the Kubernetes apps for a tenancy.
        """
        return self._tenancy_resource("kubernetes_apps", tenancy_id)


class SyncClient(BaseClient, rest.SyncClient):
    """
    Sync client for the Azimuth API.
    """
    def __enter__(self):
        obj = super().__enter__()
        obj._init()
        return self


class AsyncClient(BaseClient, rest.AsyncClient):
    """
    Async client for the Azimuth API.
    """
    async def __aenter__(self):
        obj = await super().__aenter__()
        await obj._init()
        return self
